#include "rs232.h"
#include "rs232exceptions.h"
#include "ctime.h"

CRS232::CRS232(const std::string& comm_id) : CComm(comm_id)
{
  this->serial_fd=-1;
  this->serial_device="";
  this->serial_config.baud=-1;
  this->serial_config.num_bits=-1;
  this->serial_config.parity=(parity_type)-1;
  this->serial_config.stop_bits=-1;
}

void CRS232::set_baudrate(int baud)
{
  struct termios2 config;
  int error;

  if((error=ioctl(this->serial_fd,TCGETS2,&config))<0)
  { 
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Impossible to execute the IOCTL to get the configuration structure.\n",this->comm_id);
  }
  else
  {
    config.c_cflag &= ~CBAUD;
    config.c_cflag |= BOTHER;
    config.c_ispeed = baud;
    config.c_ospeed = baud;
    this->serial_config.baud=baud;
    if((error=ioctl(this->serial_fd,TCSETS2,&config))<0)
    { 
      /* handle exceptions */
     throw CRS232Exception(_HERE_,"Impossible to execute the IOCTL to set the configuration structure.\n",this->comm_id);
    } 
  }
}

void CRS232::set_num_bits(char num_bits)
{
  struct termios2 config;
  int error,bits;

  if((error=ioctl(this->serial_fd,TCGETS2,&config))<0)
  { 
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Impossible to execute the IOCTL to get the configuration structure.\n",this->comm_id);
  }
  else
  {
    switch(num_bits)
    {
      case 5: bits=CS5;
              break;
      case 6: bits=CS6;
              break;
      case 7: bits=CS7;
              break;
      case 8: bits=CS8;
              break;
      default: /* handle exceptions */
               throw CRS232Exception(_HERE_,"Invalid number of bits per packet. See the documentation for the possible values.\n",this->comm_id);
               break;
    }
    config.c_cflag&=~CSIZE;
    config.c_cflag|=bits;
    this->serial_config.num_bits=num_bits;
    if((error=ioctl(this->serial_fd,TCSETS2,&config))<0)
    { 
      /* handle exceptions */
      throw CRS232Exception(_HERE_,"Impossible to execute the IOCTL to set the configuration structure.\n",this->comm_id);
    }
  }
}

void CRS232::set_parity(parity_type parity)
{
  struct termios2 config;
  int error;

  if((error=ioctl(this->serial_fd,TCGETS2,&config))<0)
  { 
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Impossible to execute the IOCTL to get the configuration structure.\n",this->comm_id);
  }
  else
  {
    if(parity!=none)
    {
      config.c_cflag|=PARENB;
      if(parity==odd) config.c_cflag|=PARODD;
      else if(parity==even) config.c_cflag&=~PARODD;
      else
      {
        /* handle exceptions */
        throw CRS232Exception(_HERE_,"Invalid parity type. See the documentation for the possible values.\n",this->comm_id);
      }
    }
    else config.c_cflag&=~PARENB;
    this->serial_config.parity=parity;
    if((error=ioctl(this->serial_fd,TCSETS2,&config))<0)
    { 
      /* handle exceptions */
      throw CRS232Exception(_HERE_,"Impossible to execute the IOCTL to set the configuration structure.\n",this->comm_id);
    }
  }
}

void CRS232::set_stop_bits(char stop_bits)
{
  struct termios2 config;
  int error;

  if((error=ioctl(this->serial_fd,TCGETS2,&config))<0)
  { 
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Impossible to execute the IOCTL to get the configuration structure.\n",this->comm_id);
  }
  else
  {
    if(stop_bits==2) config.c_cflag|=CSTOPB;
    else if(stop_bits==1)config.c_cflag&=~CSTOPB;
    else
    {
      /* handle execptions */
      throw CRS232Exception(_HERE_,"Invalid number of stop bits. See the documentation for the possible values.\n",this->comm_id);
    }
    this->serial_config.stop_bits=stop_bits;
    if((error=ioctl(this->serial_fd,TCSETS2,&config))<0)
    { 
      /* handle exceptions */
      throw CRS232Exception(_HERE_,"Impossible to execute the IOCTL to set the configuration structure.\n",this->comm_id);
    }
  }
}

void CRS232::hard_open(void *comm_dev)
{
  std::string *serial_dev=(std::string *)comm_dev;
  struct termios2 config;
  int error;

  if(serial_dev->size()==0)
  {
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Invalid serial port device name (empty string).\n",this->comm_id);
  }
  else
  {
    this->serial_fd=::open(serial_dev->c_str(),O_RDWR | O_NOCTTY | O_NONBLOCK | O_ASYNC);
    if(this->serial_fd==-1)
    {
      /* handle exceptions */
      throw CRS232Exception(_HERE_,"Impossible to open the serial port.\n",this->comm_id);
    }
    else
    {
      if((error=ioctl(this->serial_fd,TCGETS2,&config))<0)
      { 
        /* handle exceptions */
        throw CRS232Exception(_HERE_,"Impossible to execute the IOCTL to get the configuration structure.\n",this->comm_id);
      }
      else
      {
        config.c_cflag|=(CLOCAL | CREAD);
        config.c_cflag&=~CRTSCTS;
        config.c_lflag&=~(ICANON | ECHO | ECHOE | ISIG);
        config.c_iflag&=~(IXON | IXOFF | IXANY);
        config.c_iflag&=~(INLCR | IGNCR | ICRNL);
        config.c_oflag&=~OPOST;
        config.c_cc[VMIN]=0;
        config.c_cc[VTIME]=100;
        this->serial_device=*serial_dev;
        if((error=ioctl(this->serial_fd,TCSETS2,&config))<0)
        { 
          /* handle exceptions */
          throw CRS232Exception(_HERE_,"Impossible to execute the IOCTL to set the configuration structure.\n",this->comm_id);
        }
      }
    }
  }
}

void CRS232::hard_config(void *config)
{
  TRS232_config *serial_conf=(TRS232_config *)config;

  this->set_baudrate(serial_conf->baud);
  this->set_num_bits(serial_conf->num_bits);
  this->set_parity(serial_conf->parity);
  this->set_stop_bits(serial_conf->stop_bits);
}

int CRS232::hard_read(unsigned char *data, int len)
{
  int num_read=0;

  if((num_read=::read(this->serial_fd,data,len))==-1)
  {
    return -1;
  }

  return num_read;
}

int CRS232::hard_write(unsigned char *data, int len)
{
  int num_written=0;

  if(ioctl(this->serial_fd,TCSBRK,1)!=0) 
  {
//    static int tcsb_warned=0;
//    if (!tcsb_warned) fprintf(stderr, "TCSBRK doesn't work!\n");
//    tcsb_warned=1;
  }
  //tcdrain(this->serial_fd);
  if((num_written=::write(this->serial_fd,data,len))==-1)
  {
    return -1;
  } 

  return num_written; 
}

int CRS232::hard_get_num_data(void)
{
  int num;

  if(ioctl(this->serial_fd,FIONREAD,&num)==-1)
  {
    return -1;
  }

  return num;
}

int CRS232::hard_wait_comm_event(void)
{
  fd_set receive_set,error_set;
  int max_fd,wait_result=0;

  max_fd=this->serial_fd+1;
  FD_ZERO(&receive_set);
  FD_SET(this->serial_fd,&receive_set);
  FD_ZERO(&error_set);
  FD_SET(this->serial_fd,&error_set);
  wait_result=select(max_fd,&receive_set,NULL,&error_set,NULL);
  if(wait_result==-1)
    return -1; 
  else
  {
    if(FD_ISSET(this->serial_fd,&receive_set))/* data has been received */
    {
      return 1;
    }
    if(FD_ISSET(this->serial_fd,&error_set))
    {
      return 2;
    }
  }

  return -1;
}

void CRS232::hard_close(void)
{
  if(this->serial_fd!=-1)
  {
    ::close(this->serial_fd);
    this->serial_fd=-1;
  }
}

std::string CRS232::get_serial_device(void)
{
  return this->serial_device;
}

int CRS232::get_baudrate(void)
{
  return this->serial_config.baud;
}

int CRS232::get_num_bits(void)
{
  return this->serial_config.num_bits;
}

int CRS232::get_stop_bits(void)
{
  return this->serial_config.stop_bits;
}

parity_type CRS232::get_parity(void)
{
  return this->serial_config.parity;
}

void CRS232::set_control_signal(control_signals signal)
{
  int status;

  if(ioctl(this->serial_fd, TIOCMBIS, (const int *)&signal)==-1)
  {
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Impossible to get the control signals status.\n",this->comm_id);
  } 

  return;

  if(ioctl(this->serial_fd, TIOCMGET, &status)==-1)
  {
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Impossible to get the control signals status.\n",this->comm_id);
  } 
  status|=signal;
  if(ioctl(this->serial_fd, TIOCMSET, status)==-1)
  {
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Impossible to set the control signals status.\n",this->comm_id);
  }
}

void CRS232::clear_control_signal(control_signals signal)
{
  int status;

  if(ioctl(this->serial_fd, TIOCMBIC, (const int *)&signal)==-1)
  {
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Impossible to get the control signals status.\n",this->comm_id);
  } 

  return;

  if(ioctl(this->serial_fd, TIOCMGET, &status)==-1)
  {
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Impossible to get the control signals status.\n",this->comm_id);
  } 
  status&=~signal;
  if(ioctl(this->serial_fd, TIOCMSET, status)==-1)
  {
    /* handle exceptions */
    throw CRS232Exception(_HERE_,"Impossible to set the control signals status.\n",this->comm_id);
  }
}

bool CRS232::get_control_signal(control_signals signal)
{
  int status;

  ioctl(this->serial_fd, TIOCMGET, &status);
  if(status&signal) 
    return true;
  else
    return false;  
}

CRS232::~CRS232()
{
  this->close();
}
